/* SPDX-License-Identifier: GPL-2.0 OR BSD-3-Clause */
/*
 * Copyright (C) 2024 Intel Corporation
 */

#define MLD_DEBUGFS_OPEN_WRAPPER(name, buflen, argtype)			\
struct dbgfs_##name##_data {						\
	argtype *arg;							\
	bool read_done;							\
	ssize_t rlen;							\
	char rbuf[buflen];						\
};									\
static int _iwl_dbgfs_##name##_open(struct inode *inode,		\
				    struct file *file)			\
{									\
	struct dbgfs_##name##_data *data;				\
									\
	data = kzalloc(sizeof(*data), GFP_KERNEL);			\
	if (!data)							\
		return -ENOMEM;						\
									\
	data->read_done = false;					\
	data->arg = inode->i_private;					\
	file->private_data = data;					\
									\
	return 0;							\
}

#define MLD_DEBUGFS_READ_WRAPPER(name)					\
static ssize_t _iwl_dbgfs_##name##_read(struct file *file,		\
					char __user *user_buf,		\
					size_t count, loff_t *ppos)	\
{									\
	struct dbgfs_##name##_data *data = file->private_data;		\
									\
	if (!data->read_done) {						\
		data->read_done = true;					\
		data->rlen = iwl_dbgfs_##name##_read(data->arg,		\
						     sizeof(data->rbuf),\
						     data->rbuf);	\
	}								\
									\
	if (data->rlen < 0)						\
		return data->rlen;					\
	return simple_read_from_buffer(user_buf, count, ppos,		\
				       data->rbuf, data->rlen);		\
}

#define MLD_DEBUGFS_WRITE_WRAPPER(name, buflen, argtype)		\
static ssize_t _iwl_dbgfs_##name##_write(struct file *file,		\
					 const char __user *user_buf,	\
					 size_t count, loff_t *ppos)	\
{									\
	argtype *arg =							\
		((struct dbgfs_##name##_data *)file->private_data)->arg;\
	char buf[buflen] = {};						\
	size_t buf_size = min(count, sizeof(buf) -  1);			\
									\
	if (copy_from_user(buf, user_buf, buf_size))			\
		return -EFAULT;						\
	if (*ppos)							\
		return -EINVAL;						\
									\
	return iwl_dbgfs_##name##_write(arg, buf, buf_size);		\
}

static int _iwl_dbgfs_release(struct inode *inode, struct file *file)
{
	kfree(file->private_data);
	return 0;
}

#define _MLD_DEBUGFS_READ_FILE_OPS(name, buflen, argtype)		\
MLD_DEBUGFS_OPEN_WRAPPER(name, buflen, argtype)				\
MLD_DEBUGFS_READ_WRAPPER(name)						\
static const struct file_operations iwl_dbgfs_##name##_ops = {		\
	.read = _iwl_dbgfs_##name##_read,				\
	.open = _iwl_dbgfs_##name##_open,				\
	.llseek = generic_file_llseek,					\
	.release = _iwl_dbgfs_release,					\
}

#define _MLD_DEBUGFS_WRITE_FILE_OPS(name, buflen, argtype)		\
MLD_DEBUGFS_OPEN_WRAPPER(name, buflen, argtype)				\
MLD_DEBUGFS_WRITE_WRAPPER(name, buflen, argtype)			\
static const struct file_operations iwl_dbgfs_##name##_ops = {		\
	.write = _iwl_dbgfs_##name##_write,				\
	.open = _iwl_dbgfs_##name##_open,				\
	.llseek = generic_file_llseek,					\
	.release = _iwl_dbgfs_release,					\
}

#define _MLD_DEBUGFS_READ_WRITE_FILE_OPS(name, buflen, argtype)		\
MLD_DEBUGFS_OPEN_WRAPPER(name, buflen, argtype)				\
MLD_DEBUGFS_WRITE_WRAPPER(name, buflen, argtype)			\
MLD_DEBUGFS_READ_WRAPPER(name)						\
static const struct file_operations iwl_dbgfs_##name##_ops = {		\
	.write = _iwl_dbgfs_##name##_write,				\
	.read = _iwl_dbgfs_##name##_read,				\
	.open = _iwl_dbgfs_##name##_open,				\
	.llseek = generic_file_llseek,					\
	.release = _iwl_dbgfs_release,					\
}

#define WIPHY_DEBUGFS_HANDLER_WRAPPER(name)				\
static ssize_t iwl_dbgfs_##name##_write_handler(struct wiphy *wiphy,	\
				       struct file *file, char *buf,	\
				       size_t count, void *data)	\
{									\
	struct ieee80211_hw *hw = wiphy_to_ieee80211_hw(wiphy);		\
	struct iwl_mld *mld = IWL_MAC80211_GET_MLD(hw);			\
	return iwl_dbgfs_##name##_write(mld, buf, count, data);		\
}

#define WIPHY_DEBUGFS_WRITE_WRAPPER(name, bufsz, objtype, mldtype)	\
WIPHY_DEBUGFS_HANDLER_WRAPPER(name)					\
static ssize_t __iwl_dbgfs_##name##_write(struct file *file,		\
					  const char __user *user_buf,	\
					  size_t count, loff_t *ppos)	\
{									\
	struct ieee80211_##objtype *arg = file->private_data;		\
	struct iwl_mld_##mldtype *mldobj =				\
		iwl_mld_##mldtype##_from_mac80211(arg);			\
	struct iwl_mld *mld = mldobj->mld;				\
	char buf[bufsz] = {};						\
									\
	return wiphy_locked_debugfs_write(mld->wiphy, file,		\
				buf, sizeof(buf),			\
				user_buf, count,			\
				iwl_dbgfs_##name##_write_handler,	\
				arg);					\
}

#define _WIPHY_DEBUGFS_WRITE_FILE_OPS(name, bufsz, objtype, mldtype)	\
	WIPHY_DEBUGFS_WRITE_WRAPPER(name, bufsz, objtype, mldtype)	\
	static const struct file_operations iwl_dbgfs_##name##_ops = {	\
		.write = __iwl_dbgfs_##name##_write,			\
		.open = simple_open,					\
		.llseek = generic_file_llseek,				\
	}

#define WIPHY_DEBUGFS_WRITE_FILE_OPS(name, bufsz, objtype)		\
	_WIPHY_DEBUGFS_WRITE_FILE_OPS(name, bufsz, objtype, objtype)

#define WIPHY_DEBUGFS_HANDLER_WRAPPER_MLD(name)				\
static ssize_t iwl_dbgfs_##name##_write_handler(struct wiphy *wiphy,	\
				       struct file *file, char *buf,	\
				       size_t count, void *data)	\
{									\
	struct ieee80211_hw *hw = wiphy_to_ieee80211_hw(wiphy);		\
	struct iwl_mld *mld = IWL_MAC80211_GET_MLD(hw);			\
	return iwl_dbgfs_##name##_write(mld, buf, count);		\
}

#define WIPHY_DEBUGFS_WRITE_WRAPPER_MLD(name, bufsz)			\
WIPHY_DEBUGFS_HANDLER_WRAPPER_MLD(name)					\
static ssize_t __iwl_dbgfs_##name##_write(struct file *file,		\
					  const char __user *user_buf,	\
					  size_t count, loff_t *ppos)	\
{									\
	struct iwl_mld *mld = file->private_data;			\
	char buf[bufsz] = {};						\
									\
	return wiphy_locked_debugfs_write(mld->wiphy, file,		\
				buf, sizeof(buf),			\
				user_buf, count,			\
				iwl_dbgfs_##name##_write_handler,	\
				NULL);					\
}

#define WIPHY_DEBUGFS_WRITE_FILE_OPS_MLD(name, bufsz)			\
	WIPHY_DEBUGFS_WRITE_WRAPPER_MLD(name, bufsz)			\
	static const struct file_operations iwl_dbgfs_##name##_ops = {	\
		.write = __iwl_dbgfs_##name##_write,			\
		.open = simple_open,					\
	}
